@@ -59,23 +59,27 @@ static CoglUserDataKey offscreen_wrapper_key;
 /* This uniform is used to flip the rendering or not depending on
  * whether we are rendering to an offscreen buffer or not */
 #define MAIN_WRAPPER_FLIP_UNIFORM "_cogl_flip_vector"
+/* These comments are used to delimit the added wrapper snippet so
+ * that we can remove it again when the shader source is requested via
+ * glGetShaderSource */
+#define MAIN_WRAPPER_BEGIN "/*_COGL_WRAPPER_BEGIN*/"
+#define MAIN_WRAPPER_END "/*_COGL_WRAPPER_END*/"
 
-/* This wrapper function around 'main' is appended to every program in
- * a separate shader so that we can add some extra code to flip the
- * rendering when rendering to an offscreen buffer */
+/* This wrapper function around 'main' is appended to every vertex shader
+ * so that we can add some extra code to flip the rendering when
+ * rendering to an offscreen buffer */
 static const char
 main_wrapper_function[] =
+  MAIN_WRAPPER_BEGIN "\n"
   "uniform vec4 " MAIN_WRAPPER_FLIP_UNIFORM ";\n"
   "\n"
   "void\n"
-  MAIN_WRAPPER_REPLACEMENT_NAME " ();\n"
-  "\n"
-  "void\n"
   "main ()\n"
   "{\n"
   "  " MAIN_WRAPPER_REPLACEMENT_NAME " ();\n"
   "  gl_Position *= " MAIN_WRAPPER_FLIP_UNIFORM ";\n"
-  "}\n";
+  "}\n"
+  MAIN_WRAPPER_END;
 
 enum {
   RESTORE_FB_NONE,
@@ -732,8 +736,6 @@ gl_create_program_wrapper (void)
       g_hash_table_insert (gles2_ctx->program_map,
                            GINT_TO_POINTER (id),
                            data);
-
-      gles2_ctx->context->glAttachShader (id, gles2_ctx->wrapper_shader);
     }
 
   return id;
@@ -828,7 +830,8 @@ gl_shader_source_wrapper (GLuint shader,
                                           GINT_TO_POINTER (shader))) &&
       shader_data->type == GL_VERTEX_SHADER)
     {
-      char **string_copy = g_alloca (count * sizeof (char *));
+      char **string_copy = g_alloca ((count + 1) * sizeof (char *));
+      int *length_copy = g_alloca ((count + 1) * sizeof (int));
       int i;
 
       /* Replace any occurences of the symbol 'main' with a different
@@ -837,7 +840,12 @@ gl_shader_source_wrapper (GLuint shader,
 
       for (i = 0; i < count; i++)
         {
-          int string_length = length ? length[i] : strlen (string[i]);
+          int string_length;
+
+          if (length == NULL || length[i] < 0)
+            string_length = strlen (string[i]);
+          else
+            string_length = length[i];
 
           string_copy[i] = g_memdup (string[i], string_length);
 
@@ -845,13 +853,20 @@ gl_shader_source_wrapper (GLuint shader,
                          "main",
                          MAIN_WRAPPER_REPLACEMENT_NAME,
                          string_length);
+
+          length_copy[i] = string_length;
         }
 
-      gles2_ctx->context->glShaderSource (shader,
-                                          count,
-                                          (const char *const *) string_copy,
-                                          length);
+      string_copy[count] = (char *) main_wrapper_function;
+      length_copy[count] = sizeof (main_wrapper_function) - 1;
 
+      gles2_ctx->context->glShaderSource (shader,
+                                          count + 1,
+                                          (const char *const *) string_copy,
+                                          length_copy);
+
+      /* Note: we don't need to free the last entry in string_copy[]
+       * because it is our static wrapper string... */
       for (i = 0; i < count; i++)
         g_free (string_copy[i]);
     }
@@ -879,6 +894,23 @@ gl_get_shader_source_wrapper (GLuint shader,
       shader_data->type == GL_VERTEX_SHADER)
     {
       GLsizei copy_length = MIN (length, buf_size - 1);
+      static const char wrapper_marker[] = MAIN_WRAPPER_BEGIN;
+      char *wrapper_start;
+
+      /* Strip out the wrapper snippet we added when the source was
+       * specified */
+      wrapper_start = _cogl_util_memmem (source,
+                                         copy_length,
+                                         wrapper_marker,
+                                         sizeof (wrapper_marker) - 1);
+      if (wrapper_start)
+        {
+          length = wrapper_start - source;
+          copy_length = length;
+          *wrapper_start = '\0';
+        }
+
+      /* Correct the name of the main function back to its original */
       replace_token (source,
                      MAIN_WRAPPER_REPLACEMENT_NAME,
                      "main",
@@ -933,37 +965,6 @@ gl_get_program_iv_wrapper (GLuint program,
     }
 }
 
-static void
-gl_get_attached_shaders_wrapper (GLuint program,
-                                 GLsizei max_count,
-                                 GLsizei *count_ret,
-                                 GLuint *obj)
-{
-  CoglGLES2Context *gles2_ctx = current_gles2_context;
-  GLuint *tmp_buf;
-  GLsizei count, count_out;
-  int i;
-
-  /* We need to remove the wrapper shader we added from this list */
-
-  /* Allocate a temporary buffer that is one larger than the buffer
-   * passed in in case the application allocated exactly the size
-   * returned by GL_ATTACHED_SHADERS. */
-  tmp_buf = g_alloca (sizeof (GLuint) * (max_count + 1));
-
-  gles2_ctx->context->glGetAttachedShaders (program,
-                                            max_count + 1,
-                                            &count,
-                                            tmp_buf);
-
-  for (i = 0, count_out = 0; i < count; i++)
-    if (tmp_buf[i] != gles2_ctx->wrapper_shader)
-      obj[count_out++] = tmp_buf[i];
-
-  if (count_ret)
-    *count_ret = count_out;
-}
-
 static void
 flush_viewport_state (CoglGLES2Context *gles2_ctx)
 {
@@ -1477,8 +1478,6 @@ _cogl_gles2_context_free (CoglGLES2Context *gles2_context)
   const CoglWinsysVtable *winsys;
   GList *objects, *l;
 
-  ctx->glDeleteShader (gles2_context->wrapper_shader);
-
   if (gles2_context->current_program)
     program_data_unref (gles2_context->current_program);
 
@@ -1528,8 +1527,6 @@ _cogl_gles2_context_free (CoglGLES2Context *gles2_context)
                                  NULL);
     }
 
-  cogl_object_unref (gles2_context->context);
-
   g_free (gles2_context->vtable);
 
   g_free (gles2_context);
@@ -1557,33 +1554,6 @@ free_texture_object_data (CoglGLES2TextureObjectData *data)
   g_slice_free (CoglGLES2TextureObjectData, data);
 }
 
-static GLuint
-create_wrapper_shader (CoglContext *ctx)
-{
-  const char *strings = main_wrapper_function;
-  GLint length = sizeof (main_wrapper_function) - 1;
-  GLint status;
-  GLuint shader;
-
-  shader = ctx->glCreateShader (GL_VERTEX_SHADER);
-  ctx->glShaderSource (shader, 1, &strings, &length);
-  ctx->glCompileShader (shader);
-  ctx->glGetShaderiv (shader, GL_COMPILE_STATUS, &status);
-
-  if (!status)
-    {
-      char buf[512];
-
-      ctx->glGetShaderInfoLog (shader,
-                               sizeof (buf),
-                               NULL, /* length */
-                               buf);
-      g_warning ("Compiling wrapper shader failed:\n%s", buf);
-    }
-
-  return shader;
-}
-
 CoglGLES2Context *
 cogl_gles2_context_new (CoglContext *ctx, GError **error)
 {
@@ -1601,7 +1571,6 @@ cogl_gles2_context_new (CoglContext *ctx, GError **error)
 
   gles2_ctx = g_malloc0 (sizeof (CoglGLES2Context));
 
-  cogl_object_ref (ctx);
   gles2_ctx->context = ctx;
 
   COGL_LIST_INIT (&gles2_ctx->foreign_offscreens);
@@ -1610,7 +1579,6 @@ cogl_gles2_context_new (CoglContext *ctx, GError **error)
   gles2_ctx->winsys = winsys->context_create_gles2_context (ctx, error);
   if (gles2_ctx->winsys == NULL)
     {
-      cogl_object_unref (gles2_ctx->context);
       g_free (gles2_ctx);
       return NULL;
     }
@@ -1639,8 +1607,6 @@ cogl_gles2_context_new (CoglContext *ctx, GError **error)
 #undef COGL_EXT_FUNCTION
 #undef COGL_EXT_END
 
-  gles2_ctx->wrapper_shader = create_wrapper_shader (ctx);
-
   gles2_ctx->vtable->glBindFramebuffer = gl_bind_framebuffer_wrapper;
   gles2_ctx->vtable->glReadPixels = gl_read_pixels_wrapper;
   gles2_ctx->vtable->glCopyTexImage2D = gl_copy_tex_image_2d_wrapper;
@@ -1656,7 +1622,6 @@ cogl_gles2_context_new (CoglContext *ctx, GError **error)
   gles2_ctx->vtable->glGetShaderSource = gl_get_shader_source_wrapper;
   gles2_ctx->vtable->glLinkProgram = gl_link_program_wrapper;
   gles2_ctx->vtable->glGetProgramiv = gl_get_program_iv_wrapper;
-  gles2_ctx->vtable->glGetAttachedShaders = gl_get_attached_shaders_wrapper;
   gles2_ctx->vtable->glGetProgramInfoLog = gl_get_program_info_log_wrapper;
   gles2_ctx->vtable->glGetShaderInfoLog = gl_get_shader_info_log_wrapper;
   gles2_ctx->vtable->glClear = gl_clear_wrapper;
